==> mutual_credit-5.0.x-dev.tar.gz: mutual_credit/src/Element/WorthForm.php <==
<?php

namespace Drupal\mcapi\Element;

use Drupal\mcapi\Entity\Currency;
use Drupal\mcapi\Mcapi;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element\FormElement;

/**
 * Provides a worth field form element.
 *
 * @FormElement("worth_form")
 */
class WorthForm extends FormElement {
  use \Drupal\Core\Render\Element\CompositeFormElementTrait;

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = get_class($this);
    return [
      '#process' => [
        [$class, 'process'],
      ],
      '#element_validate' => [
        [$class, 'validate'],
      ],
      '#pre_render' => [
        //[$class, 'preRenderCompositeFormElement'],
      ],
      '#attached' => ['library' => ['mcapi/mcapi.worth.element']],
      '#attributes' => ['class' => ['worth']],
      '#config' => FALSE,
      '#minus' => FALSE, // TRUE if negative values allowed (when #config is TRUE only)
      '#tree' => TRUE,
      '#allowed_curr_ids' => array_keys(Currency::loadMultiple()),
      '#oneofmany' => FALSE,
    ];
  }

  /**
   * Form element processor callback for 'worth'.
   *
   * Takes the raw #default_values and makes child widgets according to the
   * currency format receives something like this:
   * [
   * '#type' => 'worth_form',
   * '#default_value' => [
   *   'curr_id',
   *   'value'
   * ],
   * '#allowed_curr_ids' => [cc],
   * '#config' => $element['#config'],
   * '#max' => NULL,
   * '#min' => 0,
   * ];.
   */
  public static function process($element, FormStateInterface $form_state, $form) {
    // Render the widget for a known currency
    if (count($element['#allowed_curr_ids']) == 1) {
      $currency = Currency::load(reset($element['#allowed_curr_ids']));
      if (!$currency) {
        trigger_error('Problem with allowed_currency_ids in WorthForm', E_USER_ERROR);
      }
      Self::subfields($element, $currency);
      $element['#prefix'] = '<div class="worth-widget">';
      $element['#suffix'] = '</div>';
    }
    // Render the widget for a selection of currencies.
    else {
      $element['curr_id'] = [
        '#type' => 'select',
        '#options' => array_intersect_key(
          Mcapi::entityOptions('mcapi_currency'),
          Currency::loadMultiple($element['#allowed_curr_ids'] ?: NULL)
        ),
        '#required' => TRUE,
      ];
      $element['value'] = [
        '#type' => 'number',
        '#default_value' => $element['#default_value'],
      ];
      if ($element['#config']) {
        $element['value']['#size'] = 10;
        $element['value']['#maxlength'] = 10;
      }
    }
    return $element;
  }

  /**
   * Build the widget for a single currency.
   *
   * @param array $element
   *   An unprocessed part of a render array.
   * @param Currency $currency
   *   A currency entity.
   */
  public static function subfields(array &$element, Currency $currency) {
    $element['curr_id'] = [
      '#type' => 'hidden',
      '#value'  => $currency->id(),
    ];
    $value = $element['#default_value'] ? $element['#default_value']['value'] : 0;
    if ($element['#config'] && (!is_numeric($value) || $value == 0)) {
      $value_parts = ['main' => 0, 'subdivision' => 0];
    }
    else {
      $value_parts = (strlen($value)) ?
        $currency->formattedParts($value) :
        ['main' => 0, 'subdivision' => 0];
    }
    $max_chars = strlen($currency->format['main']);
    $dec_places = strpos($currency->format['main'], '.') ?
      strlen($currency->format['main']) - strpos($currency->format['main'], '.')-1 :
      0;

    $element['main'] = [
      '#type' => 'number',
      '#default_value' => $value_parts['main']??0,
      '#min' => $element['#minus'] ? NULL : 0,
      '#step' => $dec_places ? number_format(pow(10, -$dec_places), $dec_places) : 1,
      '#size' => $max_chars + ($element['#minus'] ? 1 : 0),
      '#max' => $currency->format['main'], // Which is all the nines,
      '#min' => 0,
      '#field_prefix' => $currency->format['f1'],
      '#field_suffix' => $currency->format['f2']??'',
      '#weight' => 1
    ];
    if (isset($currency->format['subdivision'])) {
      $subdivision = $currency->format['subdivision'];
      // We need to make a dropdown.
      if (strpos($subdivision, '/')) {
        $options = [];
        list($subdivision, $divisor) = explode('/', $subdivision);
        $base = $subdivision + 1;
        $step = $base / $divisor;
        for ($j = 0; $j < $divisor; $j++) {
          $v = $j * $step;
          $options[intval($v)] = $v;
        }
        // If a preset value isn't in the $options
        // then we ignore the options and use the numeric sub-widget.
        if (isset($value_parts['subdivision']) and array_key_exists($value_parts['subdivision'], $options)) {
          $element['subdivision'] = [
            '#type' => 'select',
            '#options' => $options,
            '#attributes' => ['style' => 'width: inherit;'],
            '#weight' => 3
          ];
        }
        else {
          $size = strlen($subdivision);
          $element['subdivision'] = [
            '#type' => 'number',
            '#min' => $element['#minus'] ? NULL : 0,
            '#step' => $step,
            '#size' => $size,
            '#maxlength' => $size,
            '#attributes' => ['style' => 'width:' . ($size + 1) . 'en;'],
            '#max' => $subdivision //Which is already all the nines
          ];
          if ($element['#config']) {
            // Placeholders only work in config fields.
            $element['#type'] = 'textfield';
            if (isset($element['#placeholder'])) {
              $placeholder_val = @$element['#placeholder'];
              $p_parts = $currency->formattedParts(abs(intval($placeholder_val)));
              $element['main']['#placeholder'] = $p_parts['main'];
              if (isset($p_parts['subdivision'])) {
                $element['subdivision']['#placeholder'] = $p_parts['subdivision'];
              }
            }
          }
          else {
            // Leave the main value field empty & put zero placeholders in other
            // fields.
            $element['subdivision']['#placeholder'] = str_pad('', $size, '0');
          }
        }
      }
      else {
        $element['subdivision'] = [
          '#type' => 'number',
          '#default_value' => $value_parts['subdivision']??0,
          '#min' => 0,
          '#step' => 1,
          // #size has no effect in opera.
          '#size' => strlen($subdivision)+2, // Spinner takes about 2 chars
          '#maxlength' => strlen($subdivision),
          '#attributes' => ['style' => 'display:inline;'],
          '#max' => $currency->format['subdivision'], // E.G. 59 minutes
          '#weight' => 3
        ];
      }
      $element['subdivision'] += [
        '#weight' => 3,
        '#default_value' => $value_parts['subdivision']??0,
        '#theme_wrappers' => [],
      ];
    }
    if (isset($currency->format['f3'])) {
      $element['f3']['#markup'] = '<span>'.$currency->format['f3'].'</span>';
      $element['f3']['#weight'] = 5;
    }

    if ($element['#config'] and $element['#minus']) {
      $element['main']['#field_prefix'] = '-'.($element['main']['#field_prefix'] ??'');
      $element['main']['#default_value'] = -$element['main']['#default_value'];
    }

    $element['#attributes']['class'][] = $element['#oneofmany'] ? 'multiple' : 'single';
  }

  /**
   * {@inheritdoc}
   *
   * @todo Stop the values of the child elements overwriting the $output here?
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state) {
    if ($input == FALSE or !isset($input['curr_id'])) {
      return $element['#default_value'];
    }
    $result = static::worthValue($input, (bool)$element['#config']);
    if ($element['#config'] and $element['#minus']) {
      $result['value'] = -$result['value'];
    }
    return $result;
  }


  /**
   * Validate callback.
   */
  public static function validate(&$element, FormStateInterface $form_state) {
    $val = $element['#value'];
    if (empty($val)) {
      return;
    }
    if ($element['#config']) {
      // Test the formula.
      if (empty($val['value'])) {
        return;
      }
      $result = Self::calc($val['value'], 100);
      if (!is_numeric($result)) {
        $form_state->setError($element, t('Invalid formula'));
      }
    }
    else {
      // Check for allowed zero values.
      // zero values are only accepted if the currency allows and if cardinality
      // there is only one currency in the field.
      if ($val['value']) {
        if ($val['value'] < 0 && !($element['#config'] && $element['#minus'])) {
          // This should never happen.
          $form_state->setError(
            $element['#value']['value'],
            t(
              'Negative amounts not allowed: %val',
              ['%val' => Currency::load($val['curr_id'])->format($element['#value']['value'])]
            )
          );
        }
      }
    }
  }

  /**
   * Calculate a transaction quantity based on a provided formala and quantity.
   *
   * @param string $formula
   *   Formula using [q] for base_quant. If it is just a number the number is
   *   returned as is. Otherwise [q]% should work or other variables to be
   *   determined.
   * @param int $base_value
   *   The starting value.
   *
   * @return int | NULL
   *   The calculated result.
   */
  public static function calc(string $formula, int $base_value) {
    if (is_null($base_value)) {
      return 0;
    }
    if (is_numeric($formula)) {
      return $formula;
    }
    $proportion = str_replace('%', '', $formula);
    if (empty($base_value)) {
      // @todo there is a problem this variable is not used...
      $base_quant = 0;
    }
    if (is_numeric($proportion)) {
      return $base_value * $proportion / 100;
    }
    // $formula = str_replace('/', '//', $formula);.
    $equation = str_replace('[q]', $base_value, $formula) . ';';
    $val = eval('return ' . $equation);
    if (is_numeric($val)) {
      return $val;
    }
    $this->messenger()->addStatus(t('Problem with calculation for dependent transaction: @val', array('@val' => $val)));
  }

  /**
  *
  * @param array $input
  * @param bool $config
  * @return type
  */
  static function worthValue(array $input, bool $config) {
    if ($config && empty($input['main']) && empty($input['subdivision'])) {
      return [];
    }
    $val = Currency::load($input['curr_id'])->unformat($input);
    return [
      'curr_id' => $input['curr_id'],
      'value' => $val,
    ];
  }

}

==> mutual_credit-5.1.x-dev.tar.gz: mutual_credit/src/Element/WorthForm.php <==
<?php

namespace Drupal\mcapi\Element;

use Drupal\mcapi\Entity\Currency;
use Drupal\mcapi\Mcapi;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element\FormElement;

/**
 * Provides a worth field form element.
 *
 * @FormElement("worth_form")
 */
class WorthForm extends FormElement {
  use \Drupal\Core\Render\Element\CompositeFormElementTrait;

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = get_class($this);
    return [
      '#process' => [
        [$class, 'process'],
      ],
      '#element_validate' => [
        [$class, 'validate'],
      ],
      '#pre_render' => [
        //[$class, 'preRenderCompositeFormElement'],
      ],
      '#attached' => ['library' => ['mcapi/worth.element']],
      '#attributes' => ['class' => ['worth']],
      '#config' => FALSE,
      '#minus' => FALSE, // TRUE if negative values allowed (when #config is TRUE only)
      '#tree' => TRUE,
      '#allowed_curr_ids' => array_keys(Currency::loadMultiple()),
      '#oneofmany' => FALSE,
    ];
  }

  /**
   * Form element processor callback for 'worth'.
   *
   * Takes the raw #default_values and makes child widgets according to the
   * currency format receives something like this:
   * [
   * '#type' => 'worth_form',
   * '#default_value' => [
   *   'curr_id',
   *   'value'
   * ],
   * '#allowed_curr_ids' => [cc],
   * '#config' => $element['#config'],
   * '#max' => NULL,
   * '#min' => 0,
   * ];.
   */
  public static function process($element, FormStateInterface $form_state, $form) {
    // Render the widget for a known currency
    if (count($element['#allowed_curr_ids']) == 1) {
      $currency = Currency::load(reset($element['#allowed_curr_ids']));
      if (!$currency) {
        trigger_error('Problem with allowed_currency_ids in WorthForm', E_USER_ERROR);
      }
      Self::subfields($element, $currency);
      $element['#prefix'] = '<div class="worth-widget">';
      $element['#suffix'] = '</div>';
    }
    // Render the widget for a selection of currencies.
    else {
      $element['curr_id'] = [
        '#type' => 'select',
        '#options' => array_intersect_key(
          Mcapi::entityOptions('mcapi_currency'),
          Currency::loadMultiple($element['#allowed_curr_ids'] ?: NULL)
        ),
        '#required' => TRUE,
      ];
      $element['value'] = [
        '#type' => 'number',
        '#default_value' => $element['#default_value'],
      ];
      if ($element['#config']) {
        $element['value']['#size'] = 10;
        $element['value']['#maxlength'] = 10;
      }
    }
    return $element;
  }

  /**
   * Build the widget for a single currency.
   *
   * @param array $element
   *   An unprocessed part of a render array.
   * @param Currency $currency
   *   A currency entity.
   */
  public static function subfields(array &$element, Currency $currency) {
    $element['curr_id'] = [
      '#type' => 'hidden',
      '#value'  => $currency->id(),
    ];
    $value = $element['#default_value'] ? $element['#default_value']['value'] : 0;
    if ($element['#config'] && (!is_numeric($value) || $value == 0)) {
      $value_parts = ['main' => 0, 'subdivision' => 0];
    }
    else {
      $value_parts = (strlen($value)) ?
        $currency->formattedParts($value) :
        ['main' => 0, 'subdivision' => 0];
    }
    $max_chars = strlen($currency->format['main']);
    $dec_places = strpos($currency->format['main'], '.') ?
      strlen($currency->format['main']) - strpos($currency->format['main'], '.')-1 :
      0;

    $element['main'] = [
      '#type' => 'number',
      '#default_value' => $value_parts['main']??0,
      '#min' => $element['#minus'] ? NULL : 0,
      '#step' => $dec_places ? number_format(pow(10, -$dec_places), $dec_places) : 1,
      '#size' => $max_chars + ($element['#minus'] ? 1 : 0),
      '#max' => $currency->format['main'], // Which is all the nines,
      '#min' => 0,
      '#field_prefix' => $currency->format['f1'],
      '#field_suffix' => $currency->format['f2']??'',
      '#weight' => 1
    ];
    if (isset($currency->format['subdivision'])) {
      $subdivision = $currency->format['subdivision'];
      // We need to make a dropdown.
      if (strpos($subdivision, '/')) {
        $options = [];
        list($subdivision, $divisor) = explode('/', $subdivision);
        $base = $subdivision + 1;
        $step = $base / $divisor;
        for ($j = 0; $j < $divisor; $j++) {
          $v = $j * $step;
          $options[intval($v)] = $v;
        }
        // If a preset value isn't in the $options
        // then we ignore the options and use the numeric sub-widget.
        if (isset($value_parts['subdivision']) and array_key_exists($value_parts['subdivision'], $options)) {
          $element['subdivision'] = [
            '#type' => 'select',
            '#options' => $options,
            '#attributes' => ['style' => 'width: inherit;'],
            '#weight' => 3
          ];
        }
        else {
          $size = strlen($subdivision);
          $element['subdivision'] = [
            '#type' => 'number',
            '#min' => $element['#minus'] ? NULL : 0,
            '#step' => $step,
            '#size' => $size,
            '#maxlength' => $size,
            '#attributes' => ['style' => 'width:' . ($size + 1) . 'en;'],
            '#max' => $subdivision //Which is already all the nines
          ];
          if ($element['#config']) {
            // Placeholders only work in config fields.
            $element['#type'] = 'textfield';
            if (isset($element['#placeholder'])) {
              $placeholder_val = @$element['#placeholder'];
              $p_parts = $currency->formattedParts(abs(intval($placeholder_val)));
              $element['main']['#placeholder'] = $p_parts['main'];
              if (isset($p_parts['subdivision'])) {
                $element['subdivision']['#placeholder'] = $p_parts['subdivision'];
              }
            }
          }
          else {
            // Leave the main value field empty & put zero placeholders in other
            // fields.
            $element['subdivision']['#placeholder'] = str_pad('', $size, '0');
          }
        }
      }
      else {
        $element['subdivision'] = [
          '#type' => 'number',
          '#default_value' => $value_parts['subdivision']??0,
          '#min' => 0,
          '#step' => 1,
          // #size has no effect in opera.
          '#size' => strlen($subdivision)+2, // Spinner takes about 2 chars
          '#maxlength' => strlen($subdivision),
          '#attributes' => ['style' => 'display:inline;'],
          '#max' => $currency->format['subdivision'], // E.G. 59 minutes
          '#weight' => 3
        ];
      }
      $element['subdivision'] += [
        '#weight' => 3,
        '#default_value' => $value_parts['subdivision']??0,
        '#theme_wrappers' => [],
      ];
    }
    if (isset($currency->format['f3'])) {
      $element['f3']['#markup'] = '<span>'.$currency->format['f3'].'</span>';
      $element['f3']['#weight'] = 5;
    }

    if ($element['#config'] and $element['#minus']) {
      $element['main']['#field_prefix'] = '-'.($element['main']['#field_prefix'] ??'');
      $element['main']['#default_value'] = -$element['main']['#default_value'];
    }

    $element['#attributes']['class'][] = $element['#oneofmany'] ? 'multiple' : 'single';
  }

  /**
   * {@inheritdoc}
   *
   * @todo Stop the values of the child elements overwriting the $output here?
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state) {
    if ($input == FALSE or !isset($input['curr_id'])) {
      return $element['#default_value'];
    }
    $result = static::worthValue($input, (bool)$element['#config']);
    if ($element['#config'] and $element['#minus']) {
      $result['value'] = -$result['value'];
    }
    return $result;
  }


  /**
   * Validate callback.
   */
  public static function validate(&$element, FormStateInterface $form_state) {
    $val = $element['#value'];
    if (empty($val)) {
      return;
    }
    if ($element['#config']) {
      // Test the formula.
      if (empty($val['value'])) {
        return;
      }
      $result = Self::calc($val['value'], 100);
      if (!is_numeric($result)) {
        $form_state->setError($element, t('Invalid formula'));
      }
    }
    else {
      // Check for allowed zero values.
      // zero values are only accepted if the currency allows and if cardinality
      // there is only one currency in the field.
      if ($val['value']) {
        if ($val['value'] < 0 && !($element['#config'] && $element['#minus'])) {
          // This should never happen.
          $form_state->setError(
            $element['#value']['value'],
            t(
              'Negative amounts not allowed: %val',
              ['%val' => Currency::load($val['curr_id'])->format($element['#value']['value'])]
            )
          );
        }
      }
    }
  }

  /**
   * Calculate a transaction quantity based on a provided formala and quantity.
   *
   * @param string $formula
   *   Formula using [q] for base_quant. If it is just a number the number is
   *   returned as is. Otherwise [q]% should work or other variables to be
   *   determined.
   * @param int $base_value
   *   The starting value.
   *
   * @return int | NULL
   *   The calculated result.
   */
  public static function calc(string $formula, int $base_value) {
    if (is_null($base_value)) {
      return 0;
    }
    if (is_numeric($formula)) {
      return $formula;
    }
    $proportion = str_replace('%', '', $formula);
    if (empty($base_value)) {
      // @todo there is a problem this variable is not used...
      $base_quant = 0;
    }
    if (is_numeric($proportion)) {
      return $base_value * $proportion / 100;
    }
    // $formula = str_replace('/', '//', $formula);.
    $equation = str_replace('[q]', $base_value, $formula) . ';';
    $val = eval('return ' . $equation);
    if (is_numeric($val)) {
      return $val;
    }
    $this->messenger()->addStatus(t('Problem with calculation for dependent transaction: @val', array('@val' => $val)));
  }

  /**
  *
  * @param array $input
  * @param bool $config
  * @return type
  */
  static function worthValue(array $input, bool $config) {
    if ($config && empty($input['main']) && empty($input['subdivision'])) {
      return [];
    }
    $val = Currency::load($input['curr_id'])->unformat($input);
    return [
      'curr_id' => $input['curr_id'],
      'value' => $val,
    ];
  }

}
